# ----------------------------------------------------------------------------
# Csound Tcl-Music-Script                                          SEP-21-2002
# Copyright (c) 2002 by Hartmut Schwarz               matut@matut-spielwerk.de
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# Grundwissen der Musik fuer den Tcl-Interpreter, Lektion (1) abc-Notation
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
puts "Lektion x : abc-Notation, abc-notes";
# ----------------------------------------------------------------------------

# **********************************
# I. Beschreibung der abc-Notation *
# **********************************

# 1. Tastennummern
# ----------------

# Die Tasten einer Klaviatur werden in Halbtonschritten aufsteigend
# durchnummeriert, mit der Nummer 0 beim grossen C beginnend mit entsprechend
# negativen Werten fuer tiefere Toene. Damit ergibt sich die folgende 

# Zuordnung von Notenlinien zu Tastennummern:

# a -   21       -
# g     19
# f ----17--------
# e     16
# d ----14--------
# c     12 
# H ----11--------
# A      9  
# G -----7--------
# F      5
# E -----4--------
# D      2
# C -    0       -

# 2. Noten
# --------

# Noten sind Zeichenketten beginnend mit einem der Buchstaben C, D, E, F, G
# A, H (wahlweise auch B), c, d, e, f, g, a oder h (wahlweise auch b) zur
# Kennzeichnung des Tones. Ein dem Buchstaben folgendes + erhoeht die Note
# um einen halben Ton, ein Minuszeichen vermindert die Note um einen halben
# Ton. Eine anschliessende Folge von _ kennzeichnet mit der Gesamtlaenge der
# Note den Notenwert.

# Pausen bestehen aus einem P gefolgt von einer Folge von _ zur Kennzeichnung
# des Pausenwertes.

# 3. Notenlinien
# --------------

# Notenlinien sind aus Noten und Pausen zusammengesetzte Zeichenketten. Jeder
# Notenlinie ist eine ganze Zahl als Teiler zur Bestimmung des Notenwertes
# vorangestellt. So bedeutet eine 16, dass jedes Zeichen aus dem die Note
# besteht fuer einen Notenwert 1/16 steht. Eine Viertelnote muss dann aus 4
# Zeichen bestehen, z.B. "A+__".

# Unmittelbar hintereinander stehende Noten bilden einen Legatobereich.
# Einzeln zu spielende Noten muessen durch eine beliebige Folge beliebiger
# Zeichen getrennt werden, z.B. durch ein einzelnes Leerzeichen. Leerzeichen
# und andere nicht zu den Noten gehoerende Zeichen zaehlen nicht zum Notenwert.

# Beispiele:
# "4EGH" = "8:E_G_H_" = "16E___G___B___" ( E,G,B unter Legato-Bogen )
# "4 C E G" = "8 C_...E_...G_..." = "4 C!!!!E*53224#G"
# "8 C_C_C_C_ = "8 C______"

# 4. Metalinien
# -------------

# Metalinien sind Listen mit zusaetzlichen Informationen zur Dynamik und 
# Artikulation der einzelnen Noten. Dabei wird jeder Note der Notenlinie
# eine Zeichenkette zugewiesen. Da diese Zeichenkette wegen der Zuordnungs
# reihenfolge nicht leer sein darf, besteht sie fuer Noten ohne Metadaten
# aus einer beliebigen Anzahl von Punkten, z.B. "....".

# ----------------------------------------------------------------------------

# *******************************
# II. Die abc-Notenleseprozedur *
# *******************************
 
proc read_notes { rCTX pass_proc } {
# ----------------------------------------------------------------------------
# Liest Noten in der abc-Notation und gibt diese an die Uebergabeprozedur
# "pass_proc" weiter. Die zu lesenden Noten befinden sich im Kontext-Element
# N, die zugehoerigen Metadaten (Dynamik, Artikulation) im Kontext-Element M.
# Die Uebergabeprozedur wird mit folgenden Parametern aufgerufen:
#
#   rCTX = Context-Referenz von Leseprozedur unveraendert durchgereicht
#   indx = Notenindex oder "begin" oder "end"
#   crtv = Gesamtwert der bereits gespielten Noten
#   nval = Noten- bzw. Pausenwert der aktuell zu spielenden Note bzw. Pause
#   nkey = Tastennummer der zu spielenden Note bzw. "P" fuer Pausen
#   meta = die zur Note gehoerigen Metadaten aus der Metalinie
#   csym = Akkordsymbol
#
# Sie gibt den aktualisierten Gesamtwert der bereits gespielten Noten zurueck.
# Die Notenleseprozedur "read_notes" selbst gibt den Gesamtwert aller
# gespielten Noten zurueck.
# ----------------------------------------------------------------------------
 # --- Das Spiel vorbereiten, Initialisierungen ---
 set nkey "";  # noch keine Note gelesen
 set nval  0;  # Notenwert (time value) einer Einzelnote
 set lnk0  0;  # Legato-Bindung zum Vorgaenger 0 = nein, 1 = ja
 set lnk1  0;  # Legato-Bindung zum Nachfolger 0 = nein, 1 = ja
 set indx -1;  # Noten-Index fuer Meta-Zugriff
  
 set crtv  [clock_get_tval $rCTX]; # Aktuelle Zeit als Notenwert
 set nline [TREEget $rCTX N]; # Notenlinie, ...
 set mline [TREEget $rCTX M]; # ... Metalinie und ...
 set cline [TREEget $rCTX C]; # ... Akkordlinie aus Kontext holen

 if { [scan $nline "%d" frac] == 0 } {    # Teiler Notenwert/Zeichen holen
   set frac 16;                           # oder Standardwert 1/16 nehmen
 }
 
 # --- Anfangkennung senden ---
 set crtv [$pass_proc $rCTX "begin" $crtv 0 "P" "." "."];

 # --- Noten spielen, Schleife ueber die Notenlinie ---
 set len [string length $nline];
 for { set i 0 } { $i <= $len } { incr i } {
  if { $i < $len } {
   # --- naechstes Zeichen auf Note, Pause, Fortschreibung ... pruefen ---
   set c [string index $nline $i];
   set n [get_key_by_letter $c];
   if { ![string is double $n] } {
    switch -glob -- $c {
     "P" { set n   1 } "+" { set n  -1 } "-" { set n  -3 } "_" { set n  -2 }
     "." { set n  -4 } "?" { set n -99 }
    }
   }
  } else {
   # --- zum Schluss eine Extra-Pause um auch die letzte Note zu spielen ---
   set n 1;
  } 
  # --- Ergebnis der Pruefung auswerten ---  
  if { $n >= 0 } {
   # --- Beginn einer Note oder Pause ---
   if { [string length $nkey] > 0 } {
    # --- Metadaten fuer diese Note holen ---
    set meta  [get_meta_string $indx $mline]; 
    set csym [get_meta_string $indx $cline];
    # --- Legato-Kennung klaeren ---
    set meta [clear_legato $lnk0 $lnk1 $meta];
    # --- Die zuletzt fortgeschriebene Note spielen ---
    set nval [expr double($nval)/double($frac)];
    set crtv [$pass_proc $rCTX $indx $crtv $nval $nkey $meta $csym];
    # Legato-Bindung zum Vorgaenger fuer naechste Note aktualisieren
    set lnk0 $lnk1;
    # --- Note vergessen ---
    set nkey "";
    set nval  0;
   }
   # --- neue Note beginnen und/oder neue/alte Note fortschreiben ---
   if { $n == 1 } { # neue Pause
    set nkey "P";    
   } else {         # neue Note
    set nkey $n;
   }
   set lnk1 1;      # von Legato-Bindung zum Nachfolger ausgehen 
   incr nval;       # neue Note fortschreiben ...
   incr indx;       # ... und zaehlen
  } elseif { $n == -2 } {
   incr nval;       # alte Note fortschreiben
  } elseif { $n == -1 || $n == -3 } {
   set nkey [expr $nkey + 2 + $n]; # alte Note erhoehen/vermindern ...
   incr nval;                    # ... und fortschreiben
  } elseif { $n != -4 } { # alles Andere ausser "." ...
   set lnk1 0; # ... unterbricht Legato-Bindung zum Nachfolger
  }
 }

 # --- Endekennung senden ---
 set crtv [$pass_proc $rCTX "end" $crtv 0 "P" "." "."];

 clock_set_tval $rCTX $crtv;
 return $crtv;
} 

proc clear_legato { lnk0 lnk1 meta } {
# weil in der abc-Notation keine Legato-Kennungen ( _ ) in der Metalinie
# stehen sollen, werden diese hier eingefuegt bzw. entfernt, wenn doch
# welche drin sind.
  regsub -all ((_)|(\\\.)) $meta "" meta;  
  if { $lnk0 != $lnk1 } { # bei Legatowechsel, Kennung "_" anhaengen.
   set meta "$meta\_";
  }
  if { $meta == "" } {
   set meta ".";
  }
  return $meta;
}