#!/usr/bin/tclsh
########################################################################
# Copyright (c) 2004 by Hartmut Schwarz, matut@matut-spielwerk.de      #
#                                                                      #
# This program is free software; you can redistribute it and/or        #
# modify it under the terms of the GNU General Public License          #
# as published by the Free Software Foundation; either version 2       #
# of the License, or (at your option) any later version.               #
#                                                                      #
# This program is distributed in the hope that it will be useful,      #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         #
# GNU General Public License for more details.                         #
#                                                                      #
# You should have received a copy of the GNU General Public            #
# License along with this program; if not, write to the Free           #
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,       #
# MA 02111-1307, USA.                                                  #
########################################################################

########################################################################
# check csound log files for samples out of range and more ...         #
########################################################################

########################################################################
# function definition section                                          #
########################################################################

# load common functions
source various.tcl;

# print a minimum of help information
proc printhelp { name } {
 puts "$name : check csound log files for samples out of range and more ...";
 puts "usage : $name dirlog csndlog";
 puts "dirlog  = directory where to find csound log files";
 puts "csndlog = additional csound log file extension preceding .log";
 puts "NOTE: directory name without terminating separator";
}

# check function for a single log file
proc checklog { fpath rmaxpeak rmaxtrack rminpeak rmintrack} {
 upvar $rmaxpeak maxpeak $rmaxtrack maxtrack $rminpeak minpeak $rmintrack mintrack; 
 set status 0;
 set onum1 "none";
 set onum2 "none";
 set peak1 "none"; 
 set peak2 "none"; 
 set track [lindex [split [file tail $fpath] .] 0];
 # find samples out of range and peak amplitudes (left and right)
 set inlog [open $fpath "r"];
 while { [gets $inlog line] != -1 } {
  if { [regexp {overall samples out of range:} $line] } {
   scan $line " overall samples out of range: %d %d" onum1 onum2;
   continue;
  }
  if { [regexp {peak CH 1:} $line] } {
   scan $line "peak CH 1: %f" peak1;
   continue;
  }
  if { [regexp {peak CH 2:} $line] } {
   scan $line "peak CH 2: %f" peak2;
   continue;
  }
 }
 close $inlog;
 # handle results
 if { $onum1 == "none" || $onum2 == "none" } {
  puts "\[ERROR\] out of range info not found in track \"$track\"";
  set status 1;
 } elseif { $onum1 != 0 || $onum2 != 0 } {
  puts "\[ERROR\] $onum1 and $onum2 samples out of range found in track \"$track\"";
  set status 2;
 }
 if { $peak1 == "none" || $peak2 == "none" } {
  puts "\[ERROR\] peak info not found for track \"$track\"";
  set status 1;
 } else {
  if { $peak1 > $maxpeak } {
   set maxpeak $peak1;
   set maxtrack $track;
  }
  if { $peak2 > $maxpeak } {
   set maxpeak $peak2;
   set maxtrack $track;
  }
  if { $peak1 < $minpeak } {
   set minpeak $peak1;
   set mintrack $track;
  }
  if { $peak2 < $minpeak } {
   set minpeak $peak2;
   set mintrack $track;
  }
 }
 return $status;
}

########################################################################
# program execution section                                            #
########################################################################

# check input
if { $argc != 2 } {
 printhelp $argv0;
 exit 1;
}

# get input parameters
set dirlog  [file normalize [lindex $argv 0]];
set csndlog [lindex $argv 1];

# check logfiles
set status 0;
set maxpeak 0;
set maxtrack "none";
set minpeak 1000000;
set mintrack "none";
foreach f [glob $dirlog/*.$csndlog.log] {
 if { [checklog $f maxpeak maxtrack minpeak mintrack] != 0 } {
  set status 1;
 }
}
puts "\[NOTE\] maximum peak value $maxpeak in track \"$maxtrack\"";
puts "\[NOTE\] minimum peak value $minpeak in track \"$mintrack\"";
if { $status != 0 } {
 puts "\[ERROR\] exit with fatal track errors";
 exit 1;
}
exit 0;
